/*
 * auto_explain.c
 */
#include "postgres.h"

#include "commands/explain.h"
#include "executor/instrument.h"
#include "miscadmin.h"
#include "utils/guc_tables.h"

PG_MODULE_MAGIC;

#define GUCNAME(name)		("explain." name)

static int		explain_log_min_duration = -1;	/* msec or -1 */
static bool		explain_log_analyze = false;
static bool		explain_log_verbose = false;

static bool		toplevel = true;
static ExecutorRun_hook_type	prev_ExecutorRun = NULL;

void	_PG_init(void);
void	_PG_fini(void);

static TupleTableSlot *explain_ExecutorRun(QueryDesc *queryDesc,
										   ScanDirection direction,
										   long count);
static void init_instrument(PlanState *node);
static bool assign_log_min_duration(int newval, bool doit, GucSource source);
static bool assign_log_analyze(bool newval, bool doit, GucSource source);
static bool assign_log_verbose(bool newval, bool doit, GucSource source);

void
_PG_init(void)
{
	DefineCustomIntVariable(
		GUCNAME("log_min_duration"),
		"Sets the minimum execution time above which plans will be logged.",
		"Zero prints all plans. -1 turns this feature off.",
		GUC_UNIT_MS, &explain_log_min_duration, -1, INT_MAX / 1000,
		PGC_USERSET, assign_log_min_duration, NULL);

	DefineCustomBoolVariable(
		GUCNAME("log_analyze"),
		"Use EXPLAIN ANALYZE for plan logging.",
		NULL, 0, &explain_log_analyze,
		PGC_USERSET, assign_log_analyze, NULL);

	DefineCustomBoolVariable(
		GUCNAME("log_verbose"),
		"Use EXPLAIN VERBOSE for plan logging.",
		NULL, 0, &explain_log_verbose,
		PGC_USERSET, assign_log_verbose, NULL);

	/* install ExecutorRun_hook */
	prev_ExecutorRun = ExecutorRun_hook;
	ExecutorRun_hook = explain_ExecutorRun;
}

void
_PG_fini(void)
{
	/* uninstall ExecutorRun_hook */
	ExecutorRun_hook = prev_ExecutorRun;
}

/* initialzie instruments in PlanStaate. */
static void
init_instrument(PlanState *node)
{
	ListCell* cell;

	if (node == NULL)
		return;

	node->instrument = InstrAlloc(1);

	init_instrument(node->lefttree);
	init_instrument(node->righttree);
	foreach(cell, node->initPlan)
		init_instrument(((SubPlanState *) lfirst(cell))->planstate);
	foreach(cell, node->subPlan)
		init_instrument(((SubPlanState *) lfirst(cell))->planstate);

	switch (nodeTag(node))
	{
		case T_AppendState:
			{
				AppendState *state = (AppendState *) node;
				int	i;
				for (i = 0; i < state->as_nplans; i++)
					init_instrument(state->appendplans[i]);
			}
			break;
		case T_BitmapAndState:
			{
				BitmapAndState *state = (BitmapAndState *) node;
				int	i;
				for (i = 0; i < state->nplans; i++)
					init_instrument(state->bitmapplans[i]);
			}
			break;
		case T_BitmapOrState:
			{
				BitmapOrState *state = (BitmapOrState *) node;
				int	i;
				for (i = 0; i < state->nplans; i++)
					init_instrument(state->bitmapplans[i]);
			}
			break;
		case T_SubqueryScanState:
			return init_instrument(((SubqueryScanState *) node)->subplan);
		default:
			break;
	}
}

TupleTableSlot *
explain_ExecutorRun(QueryDesc *queryDesc, ScanDirection direction, long count)
{
	TupleTableSlot *result;
	
	if (toplevel && explain_log_min_duration >= 0)
	{
		instr_time		starttime;
		instr_time		duration;
		double			msec;

		/* Disable our hooks temporarily during the top-level query. */
		toplevel = false;
		PG_TRY();
		{
			/* Initialize instruments if not ready */
			if (explain_log_analyze && !queryDesc->doInstrument)
			{
				EState	   *estate = queryDesc->estate;
				ListCell   *cell;
				int			i;

				queryDesc->doInstrument = true;
				estate->es_instrument = true;
				for (i = 0; i < estate->es_num_result_relations; i++)
				{
					ResultRelInfo *info = &estate->es_result_relations[i];
					info->ri_TrigInstrument =
						InstrAlloc(info->ri_TrigDesc->numtriggers);
				}
				init_instrument(queryDesc->planstate);
				foreach(cell, estate->es_subplanstates)
					init_instrument((PlanState *) lfirst(cell));
			}

			INSTR_TIME_SET_CURRENT(starttime);

			if (prev_ExecutorRun)
				result = prev_ExecutorRun(queryDesc, direction, count);
			else
				result = standard_ExecutorRun(queryDesc, direction, count);

			INSTR_TIME_SET_CURRENT(duration);
			INSTR_TIME_SUBTRACT(duration, starttime);
			msec = INSTR_TIME_GET_MILLISEC(duration);

			/* Log plan if duration is exceeded. */
			if (msec > explain_log_min_duration)
			{
				StringInfoData	buf;

				initStringInfo(&buf);
				ExplainOneResult(&buf, queryDesc,
					explain_log_analyze, explain_log_verbose);

				/* Remove last line break */
				if (buf.len > 0 && buf.data[buf.len - 1] == '\n')
				{
					buf.data[buf.len - 1] = '\0';
					buf.len--;
				}
				ereport(LOG,
						(errmsg("duration: %.3f ms  plan:\n%s",
								msec, buf.data)));

				pfree(buf.data);
			}

			toplevel = true;
		}
		PG_CATCH();
		{
			toplevel = true;
			PG_RE_THROW();
		}
		PG_END_TRY();
	}
	else
	{
		/* ignore recursive executions, that are typically function calls */
		if (prev_ExecutorRun)
			result = prev_ExecutorRun(queryDesc, direction, count);
		else
			result = standard_ExecutorRun(queryDesc, direction, count);
	}

	return result;
}

/* Emulate PGC_SUSET for custom variables. */
static bool
suset_assign(GucSource source, const char *name)
{
	if (source >= PGC_S_CLIENT && !superuser())
	{
		ereport(GUC_complaint_elevel(source),
				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
				 errmsg("permission denied to set parameter \"%s\"", name)));
		return false;
	}

	return true;
}

static bool
assign_log_min_duration(int newval, bool doit, GucSource source)
{
	return suset_assign(source, GUCNAME("log_min_duration"));
}

static bool
assign_log_analyze(bool newval, bool doit, GucSource source)
{
	return suset_assign(source, GUCNAME("log_analyze"));
}

static bool
assign_log_verbose(bool newval, bool doit, GucSource source)
{
	return suset_assign(source, GUCNAME("log_verbose"));
}
